import click
from delfino.models import AppContext

from delfino_core.commands.format import run_format
from delfino_core.commands.lint import lint
from delfino_core.commands.test import test_all
from delfino_core.commands.typecheck import typecheck
from delfino_core.config import CorePluginConfig, pass_plugin_app_context
from delfino_core.utils import execute_commands_group

_COMMANDS = [run_format, lint, typecheck, test_all]


@click.command(help="Runs all verification commands. Configured by the ``verify_commands`` setting.")
@pass_plugin_app_context
@click.pass_context
def verify_all(click_context: click.Context, app_context: AppContext[CorePluginConfig]):
    execute_commands_group("verify", click_context, app_context.plugin_config)



import os
import re
import shlex
import subprocess
from datetime import datetime, timedelta
from functools import lru_cache
from typing import Optional

import click
from click import secho
from delfino.constants import PackageManager
from delfino.decorators import pass_app_context
from delfino.execution import OnError, run
from delfino.models import AppContext
from delfino.terminal_output import print_header
from delfino.utils import ArgsType
from delfino.validation import assert_package_manager_is_known, assert_pip_package_installed

from delfino_core.backports import shlex_join
from delfino_core.commands.verify_all import verify_all
from delfino_core.config import CorePluginConfig

try:
    from git import Repo
except ImportError:
    pass


_SKIP_PATTERN = (
    "Skipped Update of Package (?P<package>[^:]+): (?P<installed>[^ ]+) installed.* (?P<available>[^ ]+) available.*"
)
_OUTDATED_PATTEN = (
    "Package '(?P<package>[^']+)' out-of-date: '(?P<installed>[^']+)' installed.* '(?P<available>[^']+)' available.*"
)
_VERSION_CONSTRAINT_CHARS = "=~<>"


@lru_cache(maxsize=1)
def _git_root():
    return (
        run(["git", "rev-parse", "--show-toplevel"], capture_output=True, on_error=OnError.EXIT).stdout.decode().strip()
    )


def _trace(args: str) -> subprocess.CompletedProcess:
    """Print the command before execution."""
    _args: ArgsType = shlex.split(args)
    print(shlex_join(_args))
    return run(_args, on_error=OnError.EXIT, capture_output=True)


def _get_branch_name(repo: "Repo", start_of_week: datetime) -> str:
    email = repo.config_reader().get_value("user", "email")
    assert email, (
        "Git config has not a commit email set. Please set it with:" "\n\tgit config --global user.email 'YOUR_EMAIL'"
    )
    assert isinstance(email, str)

    user_name = re.sub("@.*", "", email)
    assert user_name, f"No user name could be parsed from git commit email '{email}' after removing '@.*'."
    assert isinstance(user_name, str)
    user_name = re.sub("[^a-zA-Z0-9._-]", "_", user_name)

    return f"{user_name}/dependencies_rollup_{start_of_week.strftime('%Y_%m_%d')}"


def _checkout_branch(repo: "Repo", branch: str):
    if str(repo.active_branch) == branch:
        secho(f"Branch '{branch}' already exists and active.", fg="green")
    elif branch in repo.branches:  # type: ignore[operator]
        secho(f"Branch '{branch}' already exists.", fg="yellow")
        _trace(f"git checkout {branch}")
    else:
        _trace("git stash")
        if repo.active_branch != "main":
            _trace("git checkout main")
        _trace("git pull")
        _trace(f"git checkout -b {branch}")
        _trace("pipenv lock")
        _trace("pipenv sync -d")


def _print_outdated_packages_and_lock_if_changed() -> bool:
    secho("Updating packages based on version pinning. This will take a while ...", fg="yellow")
    _trace("pipenv update -d")

    secho("Checking outdated packages. This will take a while ...", fg="yellow")
    result = _trace("pipenv update --outdated")

    with open("Pipfile", encoding="utf-8") as file:
        pipfile = file.read()

    available_updates = []

    for line in result.stdout.decode().split(os.linesep) + result.stderr.decode().split(os.linesep):
        match = re.match(_SKIP_PATTERN, line) or re.match(_OUTDATED_PATTEN, line)
        if not match:
            continue

        package = match.group("package")
        installed = match.group("installed").lstrip(_VERSION_CONSTRAINT_CHARS)
        available = match.group("available").lstrip(_VERSION_CONSTRAINT_CHARS)

        # Keep only packages defined in Pipfile with a different version available
        if installed != available and re.search(f'\n"?{package}"? ', pipfile):
            available_updates.append(f"{package}: {installed} -> {available}")

    if not available_updates:
        return False

    print("\n" + "\n".join(sorted(available_updates)) + "\n")
    input(
        "\033[1;33mEdit Pipfile.lock file to update any of the dependencies above ^^^\n\n"
        "Then continue by pressing ENTER ...\033[0m"
    )

    with open("Pipfile", encoding="utf-8") as file:
        return file.read() != pipfile


def _commit_and_push(repo: "Repo", commit_message: str):
    can_update = repo.commit("HEAD").message.strip() == commit_message

    if input("\033[1;33mDo you want to commit changes now? [Y/n]: \033[0m").lower() == "y":
        if can_update:  # update existing commit
            if repo.is_dirty():
                _trace("git pull")
            _trace("git commit -a --amend -C HEAD")
        else:
            _trace("git add .")
            _trace(f"git commit -a -m '{commit_message}'")

    if input("\033[1;33mDo you want to push changes now? [Y/n]: \033[0m").lower() == "y":
        if can_update:
            _trace("git push --force-with-lease")
        else:
            _trace("git push -u origin HEAD")

        url = _link_to_open_a_pull_request(repo)
        if url:
            secho(f"\nOpen a new pull request by visiting:\n\n\t{url}\n", fg="green")


def _link_to_open_a_pull_request(repo: "Repo") -> Optional[str]:
    url = repo.remote().url
    match = re.match("git@github.com:(.*)\\.git", url)
    if not match:
        match = re.match("https://github.com/(.*)\\.git", url)

    return f"https://github.com/{match.group(1)}/pull/new/{repo.active_branch}" if match else None


@click.command()
@click.option("--retry", default=False, show_default=True, is_flag=True, help=" Retry an update after failed tests.")
@pass_app_context(CorePluginConfig)
@click.pass_context
def dependencies_update(click_context: click.Context, app_context: AppContext, retry):
    """Manages the process of updating dependencies."""
    print_header("Updating dependencies", icon="🔄")

    assert_package_manager_is_known(app_context.package_manager)
    assert (
        app_context.package_manager == PackageManager.PIPENV
    ), f"Only the '{PackageManager.PIPENV.value}' package manager is supported by this command."
    assert_pip_package_installed("gitpython")

    now = datetime.utcnow()
    start_of_week = now - timedelta(days=now.isoweekday() - 1)
    repo = Repo(_git_root())

    branch_name = _get_branch_name(repo, start_of_week)
    _checkout_branch(repo, branch_name)

    if not retry:
        while True:
            if not _print_outdated_packages_and_lock_if_changed():
                break

        secho("Running all tests to check the updated dependencies ... ", fg="yellow")

    try:
        click_context.invoke(verify_all)
    except Exception:
        secho(
            "\nOne or more checks have failed. Fix any issues above and then run:\n\n\t"
            "inv dependencies-update --retry\n",
            fg="red",
        )
        raise

    commit_message = f"Dependencies rollup: {start_of_week.strftime('%Y-%m-%d')}"
    _commit_and_push(repo, commit_message)



from logging import getLogger
from typing import Dict, cast

import click
from delfino.click_utils.command import get_root_command

from delfino_core.config import CorePluginConfig

_LOG = getLogger(__name__)


def ensure_reports_dir(config: CorePluginConfig) -> None:
    """Ensures that the reports directory exists."""
    config.reports_directory.mkdir(parents=True, exist_ok=True)


def execute_commands_group(name: str, click_context: click.Context, plugin_config: CorePluginConfig, **kwargs):
    root = get_root_command(click_context)
    option_name = f"{name}_commands"

    commands: Dict[str, click.Command] = {
        command: cast(click.Command, root.get_command(click_context, command))
        for command in root.list_commands(click_context)
    }

    for target_name in getattr(plugin_config, option_name):
        if target_name not in commands:
            _LOG.warning(
                f"Command '{target_name}' used as part of the '{option_name}' option does not exist. Skipping."
            )
            continue

        if target_name not in plugin_config.disable_commands:
            click_context.forward(commands[target_name], **kwargs)



"""Linting checks on source code."""
import logging
from functools import lru_cache
from itertools import groupby
from os import getenv
from pathlib import Path
from subprocess import PIPE
from typing import List, Optional, Tuple

import click
from delfino.decorators import files_folders_option, pass_args
from delfino.execution import OnError, run
from delfino.models import AppContext
from delfino.terminal_output import print_header, print_no_issues_found
from delfino.validation import assert_pip_package_installed, pip_package_installed

from delfino_core.backports import path_is_relative_to
from delfino_core.config import CorePluginConfig, pass_plugin_app_context
from delfino_core.utils import execute_commands_group


@click.command()
@pass_args
@files_folders_option
@pass_plugin_app_context
def lint_pydocstyle(app_context: AppContext[CorePluginConfig], passed_args: Tuple[str, ...], files_folders: Tuple[str]):
    """Run docstring linting on source code.

    Docstring linting is done via pydocstyle. The pydocstyle config can be found in the
    `pyproject.toml` file under `[tool.pydocstyle]`. This ensures compliance with PEP 257,
    with a few exceptions. Note that pylint also carries out additional documentation
    style checks.
    """
    assert_pip_package_installed("pydocstyle")

    print_header("documentation style", level=2)
    dirs = build_target_paths(app_context, files_folders, False, False)

    run(
        ["pydocstyle", *passed_args, *dirs],
        stdout=PIPE,
        on_error=OnError.ABORT,
        env_update_path={"PYTHONPATH": app_context.plugin_config.sources_directory},
    )

    print_no_issues_found()


@click.command()
@pass_args
@files_folders_option
@pass_plugin_app_context
def lint_pycodestyle(
    app_context: AppContext[CorePluginConfig], passed_args: Tuple[str, ...], files_folders: Tuple[str]
):
    """Run PEP8 checking on code.

    PEP8 checking is done via pycodestyle.

    Why pycodestyle and pylint? So far, pylint does not check against every convention in PEP8. As pylint's
    functionality grows, we should move all PEP8 checking to pylint and remove pycodestyle.
    """
    assert_pip_package_installed("pycodestyle")

    print_header("code style (PEP8)", level=2)

    dirs = build_target_paths(app_context, files_folders)

    # TODO(Radek): Implement unofficial config support in pyproject.toml by parsing it
    #  and outputting the result into a supported format?
    #  See:
    #    - https://github.com/PyCQA/pycodestyle/issues/813
    #    - https://github.com/PyCQA/pydocstyle/issues/447
    args = [
        "pycodestyle",
        "--ignore",
        "E501,W503,E231,E203,E402",
        "--exclude",
        ".svn,CVS,.bzr,.hg,.git,__pycache__,.tox,*_config_parser.py",
        *passed_args,
        *dirs,
    ]
    run(
        args,
        stdout=PIPE,
        on_error=OnError.ABORT,
        env_update_path={"PYTHONPATH": app_context.plugin_config.sources_directory},
    )
    # Ignores explained:
    # - E501: Line length is checked by PyLint
    # - W503: Disable checking of "Line break before binary operator". PEP8 recently (~2019) switched to
    #         "line break before the operator" style, so we should permit this usage.
    # - E231: "missing whitespace after ','" is a false positive. Handled by black formatter.

    print_no_issues_found()


def run_pylint(
    sources_dir: Path,
    checked_dirs: List[Path],
    pylintrc_folder: Path,
    passed_args: Tuple[str, ...],
):
    print_header(", ".join(map(str, checked_dirs)), level=3)

    run(
        ["pylint", "-j", str(cpu_count()), "--rcfile", pylintrc_folder / ".pylintrc", *passed_args, *checked_dirs],
        stdout=PIPE,
        on_error=OnError.ABORT,
        env_update_path={"PYTHONPATH": sources_dir},
    )

    print_no_issues_found()


@lru_cache(maxsize=1)
def cpu_count():
    if getenv("CI", ""):
        cpu_shares = Path("/sys/fs/cgroup/cpu/cpu.shares")
        if cpu_shares.is_file():
            return int(cpu_shares.read_text(encoding="utf-8").strip()) // 1024

    log = logging.getLogger("cpu_count")
    fallback_msg = "Number of CPUs could not be determined. Falling back to 1."

    if pip_package_installed("psutil"):
        import psutil  # pylint: disable=import-outside-toplevel

        count = psutil.cpu_count(logical=False)
        if not count:
            log.warning(fallback_msg)
            return 1
        return count

    log.warning(f"`psutil` is not installed. {fallback_msg}")
    return 1


def build_target_paths(
    app_context: AppContext[CorePluginConfig],
    files_folders: Optional[Tuple[str]] = None,
    include_tests: bool = True,
    include_commands: bool = True,
) -> List[Path]:
    if files_folders:
        return [Path(path) for path in files_folders]
    local_commands_directory = app_context.pyproject_toml.tool.delfino.local_commands_directory
    plugin_config = app_context.plugin_config
    target_paths: List[Path] = [plugin_config.sources_directory]

    if include_tests and plugin_config.tests_directory.exists():
        target_paths.append(plugin_config.tests_directory)
    if include_commands and local_commands_directory.exists():
        target_paths.append(local_commands_directory)
    return target_paths


@click.command()
@files_folders_option
@pass_args
@pass_plugin_app_context
def lint_pylint(app_context: AppContext[CorePluginConfig], passed_args: Tuple[str, ...], files_folders: Tuple[str]):
    """Run pylint on code.

    The bulk of our code conventions are enforced via pylint. The pylint config can be
    found in the `.pylintrc` file.
    """
    assert_pip_package_installed("pylint")

    print_header("pylint", level=2)
    plugin_config = app_context.plugin_config

    def get_pylintrc_folder(path: Path) -> Path:
        if plugin_config.tests_directory.exists() and path_is_relative_to(path, plugin_config.tests_directory):
            return plugin_config.tests_directory
        return app_context.project_root

    target_paths = build_target_paths(app_context, files_folders)
    grouped_paths = groupby(target_paths, get_pylintrc_folder)
    for pylintrc_folder, paths in grouped_paths:
        run_pylint(app_context.plugin_config.sources_directory, list(paths), pylintrc_folder, passed_args)


@click.command(help="Runs all linting commands. Configured by the ``lint_commands`` setting.")
@files_folders_option
@pass_plugin_app_context
@click.pass_context
def lint(click_context: click.Context, app_context: AppContext[CorePluginConfig], files_folders: Tuple[str]):
    execute_commands_group("lint", click_context, app_context.plugin_config, files_folders=files_folders)
